#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#define MDB
#include "mdb.h"

#include "sz.h"
#include "webattr.h"

#define WEB_PRIVATE
#include "web.h"

enum webflags {
	WEBF_NONE = 0x0,
	WEBF_ZEN = 0x1
};

struct web {
	unsigned char magic[2];
	enum webtype type;
	int depth;
	int flags;
	web *parent;
	web *firstkid, *lastkid;
	web *next, *prev;
	sz *data;
	webattr *attr;
};

static web	*sz2web(sz *);
static web	*str2web(char *);
static void	 webkill(web *);

web *
webnew(void) {
	web *new;
	new = malloc(sizeof(web));
	if (new) {
		new->magic[0] = (unsigned char) -1;
		new->magic[1] = WEB_MAGIC;
		new->depth = 0;
		new->data = 0;
		new->attr = 0;
		new->type = 0;
		new->parent = 0;
		new->firstkid = new->lastkid = 0;
		new->next = new->prev = 0;
	}
	return new;
}

static web *
sz2web(sz *s) {
	web *w;
	w = webnew();
	w->type = WT_TEXT;
	w->data = webencode(s);
	return w;
}

static web *
str2web(char *s) {
	web *w;
	w = webnew();
	w->type = WT_TEXT;
	w->data = webencode(s);
	return w;
}

static web *
webgetp(void *v) {
	unsigned char *u = v;
	sz *s;
	web *w = 0;

	if (*u == (unsigned char) -1) {
		if (u[1] == WEB_MAGIC) {
			w = v;
			++w->depth;
		} else if ((s = szgetp(v)) != 0) {
			w = sz2web(s);
		}
	} else {
		w = str2web(v);
	}
	return w;
}

void
webkill(web *w) {
	if (!w)
		return;
	if (w->depth--)
		return;
	webfree(w);
}

void
webfree(web *w) {
	if (!w)
		return;
	szfree(w->data);
	webattrfree(w->attr);
	free(w);
}

void
websetattr(web *w, void *v1, ...) {
	va_list ap;
	void *v;
	unsigned char *u;
	sz *s1;
	webattr *wa;

	va_start(ap, v1);

	for (v = v1; v; v = va_arg(ap, void *)) {
		u = v;

		if ((*u != (unsigned char) -1) || u[1] != WEBATTR_MAGIC) {
			s1 = szgetp(u);

			if (szschr(s1, '=')) {
				wa = webattrmake(s1);
			} else {
				v = va_arg(ap, void *);
				wa = webattrmake(s1, v);
			}

			szkill(s1);
		} else {
			wa = v;
		}

		if (w->attr) {
			webattrcat(w->attr, wa);
		} else {
			w->attr = wa;
		}
	}
}

void
webfdebug(FILE *f, web *w) {
	web *tmp;
	static int depth = 0;

	fprintf(f, "%*s%p: web type %s [%d]\n",
		depth++, "",
		(void *) w,
		(w->type >= 0 && w->type < WT_MAX) ?
			web_info[w->type].name : "unknown",
		w->type);
	if (w->attr)
		webattrfdebug(f, w->attr);
	if (w->data) {
		fprintf(f, "%*sdata: '%s'\n", depth, "", szdata(w->data));
	}
	for (tmp = w->firstkid; tmp; tmp = tmp->next) {
		webfdebug(f, tmp);
	}
	fprintf(f, "%*send %s [%d]\n",
		--depth, "",
		(w->type >= 0 && w->type < WT_MAX) ?
			web_info[w->type].name : "unknown",
		w->type);

}

web *
webpage(void *v, ...) {
	web *ret;
	va_list ap;

	va_start(ap, v);
	ret = webvmake("HTML", v, ap);
	va_end(ap);

	return ret;
}

web *
webtable(void *v, ...) {
	web *ret;
	va_list ap;

	va_start(ap, v);
	ret = webvmake("TABLE", v, ap);
	va_end(ap);

	return ret;
}

web *
webvmake(void *t, void *c, va_list ap) {
	web *w = webnew();
	sz *s = szgetp(t);
	int i;

	if (!w) {
		szkill(s);
		return 0;
	}

	if (!s) {
		w->type = WT_NONE;
	} else {
		for (i = 0; i < WT_MAX; ++i) {
			if (!szicmp(s, web_info[i].name)) {
				w->type = i;
				break;
			}
		}
	}
	szkill(s);
	if (c) {
		void *v;
		for (v = c; v; v = va_arg(ap, void *)) {
			web *w1;
			w1 = webgetp(v);
			if (w1) {
				webinsert(w, w1);
			} else {
				webattr *wa = webattrgetp(v);
				websetattr(w, wa, WEBNULL);
			}
		}
	}
	return w;
}

web *
webmake(void *t, void *c, ...) {
	web *ret;
	va_list ap;

	va_start(ap, c);
	ret = webvmake(t, c, ap);
	va_end(ap);

	return ret;
}

static void
webunkid(web *parent, web *kid) {
	web *tmp;

	if (!parent || !kid || !parent->firstkid || !parent->lastkid)
		return;

	kid->parent = 0;

	if (kid == parent->firstkid && kid == parent->lastkid) {
		parent->firstkid = parent->lastkid = NULL;
		kid->prev = kid->next = NULL;
		return;
	}

	if (kid == parent->firstkid) {
		parent->firstkid = kid->next;
		if (parent->firstkid)
			parent->firstkid->prev = NULL;
		kid->prev = NULL;
		kid->next = NULL;
		return;
	}

	if (kid == parent->lastkid) {
		parent->lastkid = kid->prev;
		if (parent->lastkid)
			parent->lastkid->next = NULL;
		kid->prev = NULL;
		kid->next = NULL;
		return;
	}

	for (tmp = parent->firstkid->next; tmp; tmp = tmp->next) {
		if (tmp == kid) {
			tmp->prev->next = tmp->next;
			tmp->next->prev = tmp->prev;
			tmp->next = tmp->prev = NULL;
			return;
		}
	}
}

void
webinsert(web *w, void *content) {
	web *c = webgetp(content);
	webtype *wt;
	webinfo *wi = &web_info[w->type];

	if (wi->contains) {
		for (wt = wi->contains; *wt; ++wt)
			if (*wt == c->type)
				break;
		if (!*wt)
			return;
	}

	if (c->parent) {
		webunkid(c->parent, c);
	}
	c->parent = w;
	if (w->lastkid) {
		w->lastkid->next = c;
		c->prev = w->lastkid;
		w->lastkid = c;
		c->next = 0;
	} else {
		w->lastkid = w->firstkid = c;
	}
}

void
webfprint(FILE *f, web *w) {
	web *tmp;

	if (!w)
		return;

	if (w->type == WT_TEXT || w->type == WT_RAW) {
		fprintf(f, "%s", szdata(w->data));
	} else {
		if (web_info[w->type].spelling) {
			fprintf(f, "<%s", web_info[w->type].spelling);
			if (w->attr)
				webattrfprint(f, w->attr);
			fprintf(f, ">\n");
		}
		for (tmp = w->firstkid; tmp; tmp = tmp->next) {
			webfprint(f, tmp);
		}
		if (w->type && !(web_info[w->type].flags & WF_NOCLOSE) &&
			web_info[w->type].spelling) {
			fprintf(f, "</%s>\n", web_info[w->type].spelling);
		}
	}
}

web *
webcat(void *v, ...) {
	web *w = webgetp(v);
	void *c;
	va_list ap;

	va_start(ap, v);
	for (c = va_arg(ap, void *); c; c = va_arg(ap, void *)) {
		web *w1, *w2;
		w1 = webgetp(c);
		if (w1->type == WT_CAT) {
			web *next;
			for (w2 = w1->firstkid; w2; w2 = next) {
				next = w2->next;
				webinsert(w, w2);
			}
		} else {
			webinsert(w, w1);
		}
	}
	va_end(ap);

	return w;
}

web *
webconcat(void *v, ...) {
	web *w;
	va_list ap;

	va_start(ap, v);
	w = webvmake("CAT", v, ap);
	va_end(ap);

	return w;
}

sz *
webencode(void *v) {
	sz *s = szgetp(v);
	sz *out;
	size_t len;
	size_t i;
	char *d;

	if (!s)
		return 0;

	len = szlen(s);
	d = szdata(s);
	out = szdup("");
	for (i = 0; i < len; ++i) {
		if (isprint(d[i])) {
			switch (d[i]) {
			case '&':
				szcat(out, "&amp;");
				break;
			case '<':
				szcat(out, "&lt;");
				break;
			case '>':
				szcat(out, "&gt;");
				break;
			default:
				szccat(out, d[i]);
				break;
			}
		}
	}
	szkill(s);
	return out;
}

web *
webraw(void *v) {
	sz *s = szgetp(v);
	web *w;

	if (!s)
		return 0;

	w = webnew();
	w->type = WT_RAW;
	w->data = s;

	return w;
}
